import random # pragma: no cover
import io # pragma: no cover
from collections import Counter # pragma: no cover
import os.path # pragma: no cover
import numpy # pragma: no cover

from ._vendorized.keras_data_utils import get_file # pragma: no cover


GITHUB = 'https://github.com/UniversalDependencies/' # pragma: no cover
ANCORA_1_4_ZIP = '{github}/{ancora}/archive/r1.4.zip'.format(
    github=GITHUB, ancora='UD_Spanish-AnCora') # pragma: no cover
EWTB_1_4_ZIP = '{github}/{ewtb}/archive/r1.4.zip'.format(
    github=GITHUB, ewtb='UD_English') # pragma: no cover


def ancora_pos_tags(): # pragma: no cover
    data_dir = get_file('UD_Spanish-AnCora-r1.4', ANCORA_1_4_ZIP,
                        unzip=True)
    train_loc = os.path.join(data_dir, 'es_ancora-ud-train.conllu')
    dev_loc = os.path.join(data_dir, 'es_ancora-ud-dev.conllu')
    return ud_pos_tags(train_loc, dev_loc)


def ewtb_pos_tags(encode_tags=False, encode_words=False): # pragma: no cover
    data_dir = get_file('UD_English-r1.4', EWTB_1_4_ZIP, unzip=True)
    train_loc = os.path.join(data_dir, 'en-ud-train.conllu')
    dev_loc = os.path.join(data_dir, 'en-ud-dev.conllu')
    return ud_pos_tags(train_loc, dev_loc,
        encode_tags=encode_tags, encode_words=encode_words)


def ud_pos_tags(train_loc, dev_loc, encode_tags=True, encode_words=True): # pragma: no cover
    train_sents = list(read_conll(train_loc))
    dev_sents = list(read_conll(dev_loc))
    tagmap = {}
    freqs = Counter()
    for words, tags in train_sents:
        for tag in tags:
            tagmap.setdefault(tag, len(tagmap))
        for word in words:
            freqs[word] += 1
    vocab = {word: i for i, (word, freq) in enumerate(freqs.most_common())
             if (freq >= 10)}

    def _encode(sents):
        X = []
        y = []
        for words, tags  in sents:
            if encode_words:
                X.append([vocab.get(word, len(vocab)) for word in words])
            else:
                X.append(words)
            if encode_tags:
                y.append([tagmap[tag] for tag in tags])
            else:
                y.append(tags)
        return zip(X, y)

    return _encode(train_sents), _encode(dev_sents), len(tagmap)


def read_conll(loc): # pragma: no cover
    n = 0
    with io.open(loc, encoding='utf8') as file_:
        sent_strs = file_.read().strip().split('\n\n')
    for sent_str in sent_strs:
        lines = [line.split() for line in sent_str.split('\n')
                 if not line.startswith('#')]
        words = []
        tags = []
        for i, pieces in enumerate(lines):
            if len(pieces) == 4:
                word, pos, head, label = pieces
            else:
                idx, word, lemma, pos1, pos, morph, head, label, _, _2 = pieces
            if '-' in idx:
                continue
            words.append(word)
            tags.append(pos)
        yield words, tags


def mnist(): # pragma: no cover
    from ._vendorized.keras_datasets import load_mnist

    # the data, shuffled and split between tran and test sets
    (X_train, y_train), (X_test, y_test) = load_mnist()

    X_train = X_train.reshape(60000, 784)
    X_test = X_test.reshape(10000, 784)
    X_train = X_train.astype('float32')
    X_test = X_test.astype('float32')

    X_train = (X_train - X_train.mean(axis=0)) / (numpy.sqrt(X_train.var(axis=0)) + 1e-12)
    X_test = (X_test - X_test.mean(axis=0)) / (numpy.sqrt(X_test.var(axis=0)) + 1e-12)

    train_data = list(zip(X_train, y_train))
    nr_train = X_train.shape[0]
    random.shuffle(train_data)
    heldout_data = train_data[:int(nr_train * 0.1)]
    train_data = train_data[len(heldout_data):]
    test_data = list(zip(X_test, y_test))
    return train_data, heldout_data, test_data


def reuters(): # pragma: no cover
    from ._vendorized.keras_datasets import load_reuters
    (X_train, y_train), (X_test, y_test) = load_reuters()
    return (X_train, y_train), (X_test, y_test)


def get_word_index(path='reuters_word_index.pkl'): # pragma: no cover
    path = get_file(path, origin='https://s3.amazonaws.com/text-datasets/reuters_word_index.pkl')
    f = open(path, 'rb')

    if sys.version_info < (3,):
        data = cPickle.load(f)
    else:
        data = cPickle.load(f, encoding='latin1')

    f.close()
    return data




from ..sparse import SparseArray


def test_init():
    sp = SparseArray(10, 40.0)
    assert sp[10] == 40.0
    assert sp[0] == 0.0
    assert sp[11] == 0.0


def test_setitem():
    sp = SparseArray(10, 40.0)
    assert sp[10] == 40.0
    assert sp[0] == 0.0
    assert sp[11] == 0.0
    sp[0] = 12.0
    assert sp[0] == 12.0
    assert sp[10] == 40.0
    sp[0] = 14.0
    assert sp[0] == 14.0
    assert sp[10] == 40.0
    sp[52] = 6.0
    assert sp[0] == 14.0
    assert sp[1] == 0.0
    assert sp[10] == 40.0
    assert sp[52] == 6.0



import pytest
import numpy
from numpy.testing import assert_allclose
from cytoolz import concat
from mock import Mock


from ...neural._classes.window_encode import MaxoutWindowEncode
from ...neural._classes.embed import Embed
from ...neural.ops import NumpyOps


@pytest.fixture
def nr_out():
    return 5


@pytest.fixture
def ndim():
    return 3


@pytest.fixture
def total_length(positions):
    return sum(len(occurs) for occurs in positions.values())


@pytest.fixture
def ops():
    return NumpyOps()


@pytest.fixture
def nV(positions):
    return max(positions.keys()) + 1


@pytest.fixture
def model(ops, nr_out, ndim, nV):
    model = MaxoutWindowEncode(
                Embed(ndim, ndim, nV), nr_out, pieces=2, window=2)
    return model


@pytest.fixture
def B(positions):
    return sum(len(occurs) for occurs in positions.values())


@pytest.fixture
def gradients_BO(model, positions, B):
    gradient = model.ops.allocate((B, model.nO)) 
    for i in range(gradient.shape[0]):
        gradient[i] -= i
    return gradient


@pytest.fixture
def O(nr_out):
    return nr_out


@pytest.fixture
def I(ndim):
    return ndim


def test_nI(model):
    assert model.nI == model.embed.nO

def test_shape_is_inferred_from_data(positions, ndim, nV):
    model = MaxoutWindowEncode(
                Embed(ndim, ndim, nV), pieces=2, window=2)
    y = model.ops.asarray([[0, 1]], dtype='i')
    with model.begin_training(positions, y):
        pass
    assert model.nO == 2


def test_predict_matches_update(B, I, O, model, positions):
    x1 = model.predict(positions)
    x2, _ = model.begin_update(positions)
    assert_allclose(x1, x2)


def test_update_shape(B, I, O, model, positions, gradients_BO):
    assert gradients_BO.shape == (B, O)
    fwd, finish_update = model.begin_update(positions)
    null_grad = finish_update(gradients_BO, sgd=None)
    assert null_grad is None


def test_embed_fine_tune_is_called(model):
    positions = {10: [0]}
    W = model.W
    W.fill(1)
    gradients_BO = numpy.zeros((1, model.nO), dtype='f') - 1.
    mock_fine_tune = Mock()
    model.embed.begin_update = replace_finish_update(
        model.embed.begin_update, mock_fine_tune)
    fwd, finish_update = model.begin_update(positions)
    finish_update(gradients_BO, sgd=None)
    mock_fine_tune.assert_called_once()


def test_embed_static(model):
    positions = {10: [0]}
    W = model.W
    W.fill(1)
    gradients_BO = numpy.zeros((1, model.nO), dtype='f') - 1.
    model.embed.begin_update = replace_finish_update(
        model.embed.begin_update, None)
    fwd, finish_update = model.begin_update(positions)
    grad = finish_update(gradients_BO, sgd=None)
    assert grad is None


def replace_finish_update(begin_update, replacement):
    def replaced(*args, **kwargs):
        X, finish_update = begin_update(*args, **kwargs)
        return X, replacement
    return replaced


def test_weights_change_fine_tune(model):
    # Replace backward pass of Embed, so that it passes through the gradient.
    model.embed.begin_update = replace_finish_update(
            model.embed.begin_update, lambda gradient, sgd=None: gradient)
 
    positions = {10: [0]}
    model.W *= 0.
    model.b *= 0.
    gradients_BO = numpy.zeros((1, model.nO), dtype='f') - 1.
    
    fwd, finish_update = model.begin_update(positions)
    grad1 = finish_update(gradients_BO, sgd=None)
    
    model.W += 1
    fwd, finish_update = model.begin_update(positions)
    grad2 = finish_update(gradients_BO, sgd=None)
    
    for val1, val2 in zip(grad1.flatten(), grad2.flatten()):
        assert val1 != val2



from __future__ import print_function, unicode_literals
import numpy
from .ops import NumpyOps, CupyOps


def get_ops(ops):
    if ops in ('numpy', 'cpu'):
        return NumpyOps()
    elif ops in ('cupy', 'gpu'):
        return CupyOps()
    else:
        raise ValueError("TODO error %s" % ops)


def mark_sentence_boundaries(sequences, drop=0.): # pragma: no cover
    '''Pad sentence sequences with EOL markers.'''
    for sequence in sequences:
        sequence.insert(0, '-EOL-')
        sequence.insert(0, '-EOL-')
        sequence.append('-EOL-')
        sequence.append('-EOL-')
    return sequences, None


def flatten_sequences(sequences, drop=0.): # pragma: no cover
    ops = NumpyOps()
    return ops.flatten(sequences), None


def partition(examples, split_size): # pragma: no cover
    examples = list(examples)
    numpy.random.shuffle(examples)
    n_docs = len(examples)
    split = int(n_docs * split_size)
    return examples[:split], examples[split:]


def minibatch(stream, batch_size=1000): # pragma: no cover
    batch = []
    for X in stream:
        batch.append(X)
        if len(batch) >= batch_size:
            yield batch
            batch = []
    if len(batch) != 0:
        yield batch



import pytest
import numpy
from numpy.testing import assert_allclose
from mock import Mock
from hypothesis import given, assume
from hypothesis.extra.numpy import arrays
from hypothesis.strategies import integers


from ...neural._classes.window_encode import MaxoutWindowEncode
from ...neural._classes.embed import Embed
from ...neural.ops import NumpyOps

from ...neural._classes.window_encode import _get_vector_gradients
from ...neural._classes.window_encode import _get_full_inputs
from ...neural._classes.window_encode import _compute_hidden_layer
from ...neural.ops import NumpyOps

from ..strategies import arrays_BOP_BO
from ..strategies import arrays_OPFI_BI_lengths
from ..strategies import lengths


@pytest.fixture
def nr_out():
    return 5


@pytest.fixture
def ndim():
    return 3


@pytest.fixture
def total_length(positions):
    return sum(len(occurs) for occurs in positions.values())


@pytest.fixture
def ops():
    return NumpyOps()


@pytest.fixture
def nV(positions):
    return max(positions.keys()) + 1


@pytest.fixture
def model(ops, nr_out, ndim, nV):
    model = MaxoutWindowEncode(
                Embed(ndim,ndim, nV), nr_out, ndim, pieces=2, window=2)
    return model


@pytest.mark.slow
@pytest.mark.xfail
def test_forward_succeeds(model, ids, positions, vectors, lengths):
    out, whiches = model._forward(positions, vectors, lengths)


@pytest.mark.slow
@pytest.mark.xfail
def test_predict_batch_succeeds(model, ids, vectors, lengths):
    ids = list(toolz.concat(ids))
    out = model.predict_batch((ids, vectors, lengths))
    assert out.shape == (sum(lengths), model.nr_out)

def test_zero_gradient_makes_zero_finetune(model):
    positions = {10: [0]}
    fwd, finish_update = model.begin_update(positions)
    gradients_BO = model.ops.allocate((1, model.nO))
    finish_update(gradients_BO, sgd=None)
    assert_allclose(model.embed.d_vectors, 0)


@pytest.mark.slow
@pytest.mark.xfail
@pytest.mark.skip
@given(arrays_BOP_BO())
def test_only_one_piece_gets_gradient_if_unique_max(x_BOP_d_BO):
    x_BOP, d_BO = x_BOP_d_BO
    assert len(x_BOP.shape) == 3
    assert x_BOP.shape[0] == d_BO.shape[0]
    assert x_BOP.shape[1] == d_BO.shape[1]
    whiches_BO = numpy.argmax(x_BOP, axis=-1)
    d_BOP = numpy.zeros(x_BOP.shape)

    _get_full_gradients(d_BOP, d_BO, whiches_BO)

    for b in range(x_BOP.shape[0]):
        for o in range(x_BOP.shape[1]):
            num_at_max = sum(x_BOP[b,o] >= x_BOP[b, o, whiches_BO[b,o]])
            if num_at_max == 1.:
                if d_BO[b, o] != 0:
                    assert sum(cell != 0. for cell in d_BOP[b, o]) == 1.
                    assert abs(d_BOP[b, o]).argmax(axis=-1) == x_BOP[b, o].argmax()
                assert sum(d_BOP[b, o]) == d_BO[b, o]


@pytest.mark.slow
@given(arrays_OPFI_BI_lengths(max_B=10, max_P=10, max_I=10))
def test_compute_hidden_layer(arrays_OPFI_BI_lengths):
    W__OPFI, vectors__UI, _ = arrays_OPFI_BI_lengths
    # Converting to int saves us from a lot of annoying false positives from
    # floating point arithmetic
    W__OPFI = numpy.asarray(W__OPFI, dtype='int32')
    vectors__UI = numpy.asarray(vectors__UI, dtype='int32')
    O, P, F, I = W__OPFI.shape
    U = vectors__UI.shape[0]
    # The function works on types, not tokens. But for now assume every token
    # is unique.
    assume(F == 5)
    ops = NumpyOps()

    H__UFOP = _compute_hidden_layer(ops, W__OPFI, vectors__UI)
    assert H__UFOP.shape == (U, F, O, P)

    for u in range(U):
        expected__OPF = numpy.tensordot(W__OPFI, vectors__UI[u], axes=[[3], [0]])
        assert expected__OPF.shape == (O, P, F)
        expected__FOP = expected__OPF.transpose((2, 0, 1))
        computed = H__UFOP[u]
        assert expected__FOP.shape == computed.shape
        assert_allclose(computed, expected__FOP)


@pytest.mark.slow
@pytest.mark.xfail
@pytest.mark.skip
@given(
    lengths().flatmap(lambda lenlen:
        arrays('int32', shape=(lenlen,),
        elements=integers(min_value=1, max_value=10))))
def test_zero_features_past_sequence_boundaries(seq_lengths):
    ops = NumpyOps()
    B = sum(seq_lengths)
    O = 1
    P = 1
    H__BFOP = ops.allocate((B, 5, O, P))
    H__BFOP += 1
    _zero_features_past_sequence_boundaries(H__BFOP, seq_lengths)
    sequences = ops.unflatten(H__BFOP, seq_lengths)
    for sequence in sequences:
        w0 = sequence[0].reshape((5, O, P))
        assert_allclose(w0[0], 0)
        assert_allclose(w0[1], 0)
        if len(sequence) >= 2:
            w1 = sequence[1].reshape((5, O, P))
            assert_allclose(w1[0], 0)
        w_m1 = sequence[-1].reshape((5, O, P))
        assert_allclose(w_m1[3], 0)
        assert_allclose(w_m1[4], 0)
        if len(sequence) >= 2:
            w_m2 = sequence[-2].reshape((5, O, P))
            assert_allclose(w_m2[4], 0)


@pytest.mark.slow
@pytest.mark.xfail
def test_get_full_inputs_zeros_edges():
    B = 11
    F = 5
    I = 3
    output = numpy.zeros((B, F, I))
    vectors = numpy.ones((B, I))
    lengths = [5, 6]
    assert sum(lengths) == B
    _get_full_inputs(output,
        vectors, lengths)
    assert_allclose(output[0, 0], 0)
    assert_allclose(output[0, 1], 0)
    assert_allclose(output[1, 0], 0)
    assert_allclose(output[1, 1], 1)
    assert_allclose(output[3, 0], 1)
    assert_allclose(output[3, 1], 1)
    assert_allclose(output[3, 3], 1)
    assert_allclose(output[3, 4], 0)
    assert_allclose(output[4, 0], 1)
    assert_allclose(output[4, 1], 1)
    assert_allclose(output[4, 3], 0)
    assert_allclose(output[4, 4], 0)
    assert_allclose(output[5, 0], 0)
    assert_allclose(output[5, 1], 0)
    assert_allclose(output[6, 0], 0)
    assert_allclose(output[6, 1], 1)
    assert_allclose(output[6, 3], 1)
    assert_allclose(output[6, 4], 1)
    assert_allclose(output[10, 0], 1)
    assert_allclose(output[10, 1], 1)
    assert_allclose(output[10, 2], 1)
    assert_allclose(output[10, 3], 0)
    assert_allclose(output[10, 4], 0)





# cython: profile=True
# cython: cdivision=True
# cython: infer_types = True
cimport cython
from libc.string cimport memcpy, memset
from libc.math cimport exp, sqrt
from libc.stdlib cimport srand, rand
from libc.stdlib cimport calloc, malloc, free
from libc.string cimport memcpy
from cymem.cymem cimport Pool

import numpy
from cytoolz import concat
from numpy import prod
from numpy cimport ndarray
from collections import Sized
cimport numpy as np

from ..typedefs cimport weight_t
from ..linalg cimport VecVec


try:
    import cupy
except ImportError:
    cupy = None

try:
    import cytoolz as toolz
except ImportError:
    import toolz



class Ops(object):
    device = 'cpu'
    xp = None

    def __init__(self, xp=None):
        if xp is not None:
            self.xp = xp

    def dropout(self, x, dropout, inplace=False):
        if dropout <= 0.0:
            return x, lambda func: func
        mask = self.get_dropout_mask(x.shape, dropout)
        if mask is None:
            return x, lambda func: func
        def wrap_backprop(backprop):
            def finish_update(gradient, *args, **kwargs):
                return backprop(gradient * mask, *args, **kwargs)
            return finish_update
        if inplace:
            x *= mask
            return x, wrap_backprop
        else:
            return x * mask, wrap_backprop

    def flatten(self, X, dtype=None):
        return self.asarray(list(concat(X)), dtype=dtype)
 
    def unflatten(self, X, lengths):
        unflat = []
        for length in lengths:
            unflat.append(X[:length])
            X = X[length:]
        assert len(X) == 0
        assert len(unflat) == len(lengths)
        return unflat

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def get_dropout_mask(self, shape, drop):
        cdef unsigned char cutoff = drop * 255
        if cutoff <= 0:
            return None
        elif cutoff >= 255:
            return self.allocate(shape)
        
        cdef int n = prod(shape)
        cdef bytes rand_bytes = self.random_bytes(n)
        cdef unsigned char* buff = <unsigned char*>rand_bytes
        cdef ndarray[float] output = self.allocate(n, dtype='float32')
        cdef float* out_buff = <float*>output.data
        cdef float compensated = 1. / (1. - drop)
        for i in range(n):
            out_buff[i] = compensated if buff[i] < cutoff else 0
        return output.reshape(shape)

    def allocate(self, shape, dtype='float32'):
        if isinstance(shape, int):
            shape = (shape,)
        nr_weight = numpy.prod(shape)
        return self.xp.zeros(shape, dtype=dtype)

    def unzip(self, data):
        X, y = zip(*data)
        return self.asarray(X), self.asarray(y)

    def asarray(self, data, dtype=None):
        return self.xp.asarray(data, dtype=dtype)

    def batch_dot(self, x, y):
        return self.xp.tensordot(x, y, axes=[[1], [1]])
   
    def batch_outer(self, x, y):
        return self.xp.tensordot(x, y, axes=[[0], [0]])

    def norm(self, x):
        return self.xp.sqrt((x * x).sum())

    def dot(self, x, y):
        return self.xp.dot(x, y)
    
    def affine(self, weights, bias, signal):
        return self.batch_dot(signal, weights) + bias

    def argmax(self, x, axis=-1):
        return self.xp.argmax(x, axis=axis)

    def softmax(self, x, inplace=False, axis=1):
        if x.ndim >= 3:
            raise NotImplementedError(
                "Softmax currently only supports 2d. ndim=%d" % x.ndim)
        shape = x.shape
        maxes = self.xp.max(x, axis=1)
        maxes = maxes.reshape((x.shape[0], 1))
        shifted = x - maxes
        new_x = self.xp.exp(shifted)
        new_x /= new_x.sum(axis=1).reshape((x.shape[0], 1))
        if inplace:
            x[:] = new_x
            return x
        else:
            return new_x

    def expand_dims(self, a, axis=-1):
        return self.xp.expand_dims(a, axis=axis)

    def clip_low(self, x, value, inplace=False):
        if inplace:
            return self.xp.maximum(x, value, out=x)
        else:
            return self.xp.maximum(x, value)

    def take_which(self, x, which, axis=-1):
        output = self.allocate(which.shape)
        for i in range(x.shape[axis]):
            output += x[:,:,i] * (which == i)
        return output

    def backprop_take(self, dX__bo, which__bo, nP):
        dX__bop = self.allocate((dX__bo.shape[0], dX__bo.shape[1], nP))
        for i in range(nP):
            dX__bop[:, :, i] += dX__bo * (which__bo == i)
        return dX__bop

    def xavier_uniform_init(self, W, inplace=True):
        scale = self.xp.sqrt(6. / (W.shape[0] + W.shape[1]))
        if inplace:
            W[:] = self.xp.random.uniform(-scale, scale, W.shape)
            return W
        else:
            return self.xp.random.uniform(-scale, scale, W.shape)

    def he_normal_init(self, shape, fan_in):
        scale = self.xp.sqrt(2. / fan_in)
        return self.xp.random.normal(scale=scale, size=prod(shape)).reshape(shape)


class NumpyOps(Ops):
    device = 'cpu'
    xp = numpy

    def elu(self, ndarray X, inplace=True):
        cdef weight_t* data = <weight_t*>X.data
        cdef size_t size = X.size
        for i in range(size):
            if data[i] < 0:
                data[i] = exp(data[i])-1.

    def backprop_elu(self, ndarray delta_, ndarray signal_out_,
            inplace=True):
        # Backprop the ELU transformation
        # Note that this is over the function _output_, not the function
        # _input_!
        cdef size_t size = delta_.size
        cdef weight_t* delta = <weight_t*>delta_.data
        cdef const weight_t* signal_out = <const weight_t*>signal_out_.data
        for i in range(size):
            if signal_out[i] <= 0:
                delta[i] *= signal_out[i] + 1.

    def relu(self, ndarray X, inplace=True):
        cdef weight_t* data = <weight_t*>X.data
        cdef size_t size = X.size
        for i in range(size):
            if data[i] < 0:
                data[i] = 0.

    def backprop_relu(self, ndarray delta_, ndarray signal_out_, inplace=True):
        cdef size_t size = delta_.size
        cdef weight_t* delta = <weight_t*>delta_.data
        cdef const weight_t* signal_out = <const weight_t*>signal_out_.data
        for i in range(size):
            if signal_out[i] <= 0:
                delta[i] = 0.

    def maxout(self, float[:, :, ::1] py_cands):
        cdef Pool mem = Pool()
        cdef int B = py_cands.shape[0]
        cdef int O = py_cands.shape[1]
        cdef int P = py_cands.shape[2]

        which__bo = <int*>mem.alloc(B * O, sizeof(int))
        best__bo = <float*>mem.alloc(B * O, sizeof(float))
        maxout(best__bo, which__bo,
            &py_cands[0, 0, 0], B, O, P)
        cdef ndarray py_best = self.xp.ascontiguousarray(self.allocate(B * O, dtype='float32'))
        memcpy(py_best.data, best__bo, B * O * sizeof(best__bo[0]))
        cdef ndarray py_which = self.xp.ascontiguousarray(self.allocate(B * O, dtype='int32'))
        memcpy(py_which.data, which__bo, B * O * sizeof(which__bo[0]))
        return py_best.reshape((B, O)), py_which.reshape((B, O))
    
    def backprop_maxout(self, float[:, ::1] dX__bo, int[:, ::1] which__bo, int P):
        cdef Pool mem = Pool()
        cdef int B = dX__bo.shape[0]
        cdef int O = dX__bo.shape[1]

        dX__bop = <float*>mem.alloc(B * O * P, sizeof(float))
        backprop_maxout(dX__bop,
            &dX__bo[0, 0], &which__bo[0, 0], B, O, P)
        cdef ndarray py_out = self.xp.ascontiguousarray(self.allocate(B*O*P, dtype='float32'))
        memcpy(py_out.data, dX__bop, B * O * P * sizeof(dX__bop[0]))
        return py_out.reshape((B, O, P))

    def seq2col(self, float[:, ::1] seq, int nW):
        '''Given an (M, N) sequence of vectors, return an (M, N*(nW*2+1)) sequence.
        The new sequence is constructed by concatenating nW preceding and succeeding
        vectors onto each column in the sequence, to extract a window of features.
        '''
        cdef int B = seq.shape[0]
        cdef int I = seq.shape[1]
        cdef Pool mem = Pool()
        cols = <float*>mem.alloc(B * I * (nW*2+1), sizeof(float))
        seq2col(cols,
            &seq[0,0], B, I, nW)
        cdef ndarray py_out = self.xp.ascontiguousarray(
            self.allocate(B*(2 * nW+1) * I, dtype='float32'))
        memcpy(py_out.data, cols, B * (2*nW+1) * I * sizeof(cols[0]))
        return py_out.reshape((B, I * (2*nW+1)))
    
    def backprop_seq2col(self, float[:, ::1] dY, int nW):
        cdef int B = dY.shape[0]
        cdef int nF = nW*2+1
        cdef int I = dY.shape[1] / nF
        cdef Pool mem = Pool()
        dX = <float*>mem.alloc(B * I, sizeof(float))
        backprop_seq2col(dX, &dY[0,0], B, I, nW)
        cdef ndarray py_out = self.xp.ascontiguousarray(
            self.allocate(B * I, dtype='float32'))
        memcpy(py_out.data, dX, B * I * sizeof(dX[0]))
        return py_out.reshape((B, I))

    def increment_slices(self, ndarray contig_array, ndarray _to_add, _starts):
        cdef ndarray contig_to_add = self.xp.ascontiguousarray(_to_add, dtype='float32')
        cdef ndarray contig_starts = self.xp.ascontiguousarray(_starts, dtype='int32')

        cdef const float* to_add = <const weight_t*>contig_to_add.data
        cdef float* whole_array = <weight_t*>contig_array.data
        cdef const int* starts = <const int*>contig_starts.data
        cdef int n_slice = len(_starts)
        cdef int length = _to_add.size
        cdef int stride = length / _to_add.shape[0]
        for start in starts[:n_slice]:
            workon = &whole_array[start * stride]
            for i in range(length):
                workon[i] += to_add[i]

    def random_bytes(self, n):
        return self.xp.random.bytes(n)

class CupyOps(Ops):
    device = 'gpu'
    xp = cupy

def seed_srand(int value):
    srand(value)

cdef void fill_random_bytes(unsigned char* out, int n) nogil:
    '''Fill an array `out` with `n` random bytes.'''
    cdef int rand_bytes
    cdef unsigned char rand_byte
    cdef int i = 0
    cdef int step = sizeof(int)
    for i from 0 <= i < n by step:
        rand_bytes = rand()
        for b in range(sizeof(rand_bytes)):
            rand_byte = (rand_bytes >> (b * 8)) & 0xFF
            out[i] = rand_byte


cdef void seq2col(float* output, const float* X, int B, int I, int nW) nogil:
    nF = nW * 2 + 1
    output += nW * I
    for i in range(B-nW):
        memcpy(output,
            X, I * (nW+1) * sizeof(output[0]))
        output += I * (nW+1)
        memcpy(output,
            X, I * nW * sizeof(output[0]))
        output += I * nW
        X += I
    memcpy(output,
        X, I * nW * sizeof(output[0]))


cdef void backprop_seq2col(float* d_seqs,
        const float* d_cols, int B, int I, int nW) nogil:
    # Here's what we're doing, if we had 2d indexing.
    #for i in range(B):
    #    d_seq[i] += d_cols[i-2, 4]
    #    d_seq[i] += d_cols[i-1, 3]
    #    d_seq[i] += d_cols[i+2, 0]
    #    d_seq[i] += d_cols[i+1, 1]
    #    d_seq[i] += d_cols[i, 2]

    nF = nW * 2 + 1
    for i in range(B):
        seq_row = &d_seqs[i * I]
        col_row = &d_cols[i * I * nF]
        for f in range(-nW, nW+1):
            if B > (i+f) >= 0:
                feat = col_row + (f * I)
                VecVec.add_i(seq_row, &feat[(f+nW) * I], 1., I)



cdef void maxout(float* best__bo, int* which__bo,
        const float* cands__bop, int B, int O, int P) nogil:
    for b in range(B):
        for o in range(O):
            which__bo[0] = 0
            best__bo[0] = cands__bop[0]
            cands__bop += 1
            for p in range(1, P):
                if cands__bop[0] > best__bo[0]:
                    which__bo[0] = p
                    best__bo[0] = cands__bop[0]
                cands__bop += 1
            best__bo += 1
            which__bo += 1


cdef void backprop_maxout(float* dX__bop,
        const float* dX__bo, const int* which__bo, int B, int O, int P) nogil:
    for b in range(B):
        for o in range(O):
            dX__bop[which__bo[0]] = dX__bo[0]
            dX__bop += P
            dX__bo += 1
            which__bo += 1



from numpy import prod
import contextlib

from .. import util
from ..train import Trainer
from ..ops import NumpyOps
from ..mem import Memory
from ..util import get_ops
from ... import check
from ... import describe
from ...check import equal_length, has_shape, is_sequence, is_float, is_array


class Model(object):
    '''Model base class.'''
    name = 'model'
    id = 0
    ops = NumpyOps()
    Trainer = Trainer
    descriptions = []
    on_data_hooks = []
    on_init_hooks = [] # Use this to add layers
    _operators = {}

    @classmethod
    @contextlib.contextmanager
    def define_operators(cls, operators):
        '''Bind operators to specified functions for the scope of the context:

        Example
        -------

            model = Model()
            other = Model()
            with Model.define_operators({"+": lambda self, other: "plus"}):
                print(model + other)
                # "plus"
            print(model + other)
            # Raises TypeError --- binding limited to scope of with block.
        '''
        old_ops = dict(cls._operators)
        for op, func in operators.items():
            cls._operators[op] = func
        yield
        cls._operators = old_ops

    @classmethod
    @contextlib.contextmanager
    def use_device(cls, device):
        '''Change the device to execute on for the scope of the block.'''
        if device == cls.ops.device:
            yield
        else:
            curr_ops = cls.ops
            cls.ops = get_ops(device)
            yield
            cls.ops = curr_ops

    @property
    def input_shape(self):
        raise NotImplementedError

    @property
    def output_shape(self):
        raise NotImplementedError

    def __init__(self, *args, **kwargs):
        Model.id += 1
        self.id = Model.id
        self.name = self.__class__.name
        kwargs = self._update_defaults(args, kwargs)
        self._mem = Memory(self.ops)
        self._dims = {}
        if not hasattr(self, '_layers'):
            self._layers = []
        self.descriptions = dict(self.descriptions)
        self.on_init_hooks = list(self.on_init_hooks)
        self.on_data_hooks = list(self.on_data_hooks)

        for attr, install in self.descriptions.items():
            install(attr, self)
        for hook in self.on_init_hooks:
            hook(self, *args, **kwargs)

    def _update_defaults(self, args, kwargs):
        new_kwargs = {}
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                new_kwargs[key] = value
        return new_kwargs

    #@check.args(equal_length)
    @check.arg(1, is_sequence)
    def begin_training(self, train_X, train_y=None):
        for hook in self.on_data_hooks:
            hook(self, train_X, train_y)
        return self.Trainer(self, train_X, train_y)

    @check.arg(2, is_float)
    @check.arg(1, has_shape(('nB', 'nI')))
    def begin_update(self, X, drop=0.0):
        raise NotImplementedError

    def predict(self, X):
        y, _ = self.begin_update(X)
        return y

    def predict_one(self, x):
        X = self.ops.expand_dims(x, axis=0)
        return self.predict(X)[0]

    @contextlib.contextmanager
    def use_params(self, params): # pragma: no cover
        backup = None
        weights = self._mem.weights
        if id(self._mem) in params:
            param = params[id(self._mem)]
            backup = weights.copy()
            weights[:] = param
        if hasattr(self, '_layers'):
            contexts = [layer.use_params(params) for layer in self._layers]
            for context in contexts:
                next(context.gen)
        yield
        if backup is not None:
            weights[:] = backup
        for context in contexts:
            next(context.gen)

    def __call__(self, x):
        '''
        x
            Must match expected type
            Must match expected shape
        '''
        return self.predict(x)

    def evaluate(self, X, y):
        '''
        x
            Must match expected type
            Must match expected shape
        y
            Must match expected type
        '''
        correct = 0
        total = 0
        scores = self(X)
        for i, gold in enumerate(y):
            correct += scores[i].argmax() == gold
            total += 1
        return float(correct) / total

    @check.operator_is_defined('+')
    def __add__(self, other):
        '''Apply the function bound to the '+' operator.'''
        return self._operators['+'](self, other)

    @check.operator_is_defined('-')
    def __sub__(self, other):
        '''Apply the function bound to the '-' operator.'''
        return self._operators['-'](self, other)

    @check.operator_is_defined('*')
    def __mul__(self, other):
        '''Apply the function bound to the '*' operator.'''
        return self._operators['*'](self, other)

    @check.operator_is_defined('@')
    def __matmul__(self, other):
        '''Apply the function bound to the '@' operator.'''
        return self._operators['@'](self, other)

    @check.operator_is_defined('/')
    def __div__(self, other):
        '''Apply the function bound to the '/' operator.'''
        return self._operators['/'](self, other)

    @check.operator_is_defined('/')
    def __truediv__(self, other): # pragma: no cover
        '''Apply the function bound to the '/' operator.'''
        return self._operators['/'](self, other)

    @check.operator_is_defined('//')
    def __floordiv__(self, other):
        '''Apply the function bound to the '//' operator.'''
        return self._operators['//'](self, other)

    @check.operator_is_defined('%')
    def __mod__(self, other):
        '''Apply the function bound to the '%' operator.'''
        return self._operators['%'](self, other)

    @check.operator_is_defined('**')
    def __pow__(self, other, modulo=None):
        '''Apply the function bound to the '**' operator.'''
        return self._operators['**'](self, other)

    @check.operator_is_defined('<<')
    def __lshift__(self, other):
        '''Apply the function bound to the '<<' operator.'''
        return self._operators['<<'](self, other)

    @check.operator_is_defined('>>')
    def __rshift__(self, other):
        '''Apply the function bound to the '>>' operator.'''
        return self._operators['>>'](self, other)

    @check.operator_is_defined('&')
    def __and__(self, other):
        '''Apply the function bound to the '&' operator.'''
        return self._operators['&'](self, other)

    @check.operator_is_defined('^')
    def __xor__(self, other):
        '''Apply the function bound to the '^' operator.'''
        return self._operators['^'](self, other)

    @check.operator_is_defined('|')
    def __or__(self, other):
        '''Apply the function bound to the '|' operator.'''
        return self._operators['|'](self, other)


##
#    def pipe(self, stream, batch_size=1000):
#        for batch in util.minibatch(stream, batch_size):
#            ys = self.predict_batch(batch)
#            for y in ys:
#                yield y
#
#    def update(self, stream, batch_size=1000):
#        for X, y in util.minibatch(stream, batch_size=batch_size):
#            output, finish_update = self.begin_update(X)
#            gradient = finish_update(y)
#            yield gradient
#
#def list_gradients(self):
#    pass
#def check_input(self, x, expect_batch=False):
#    if self.layers:
#        return self.layers[0].check_input(x, expect_batch=expect_batch)
#    if is_batch(x):
#        shape = x.shape[1:]
#    else:
#        raise ShapeError.expected_batch(locals(), globals())
#    if shape != self.input_shape:
#        raise ShapeError.dim_mismatch(self.input_shape, shape)
#    else:
#        return True
#def allocate_params(model):
#    '''Allocate all trainable parameters of a model, including for its sublayers,
#    so that parameters can be contiguous in memory.'''
#    # Get total size
#    size = sum(prod(shape) for shape in model.describe_all_params)
#    params = Params(total=size)
#    for i, layer in enumerate(all_layers(model)):
#        layer.name = '%s-%d' % (layer.name, i)
#        for name, shape, init in describe_params(model):
#            params.add(name, shape)
#            if init is not None:
#                init(params.get(name), inplace=True)
#    return params
#
#
#def set_dimensions_given_data(model, X, y):
#    pass
#
#
#def initialize_weights_given_data(model, X, y):
#    for name, weights in model.weights.items():
#        init = model.get_initialzer(name)
#
#    for key, (name, shape, init_func) in model.description.weights:
#        if init_func is not None:
#            weights = model.w.get(key)
#            init_func(weights, X, y)
#
#
#
#def all_layers(model):
#    '''Iterate over all layers in the model.'''
#    queue = [model]
#    seen = set()
#    for layer in queue:
#        yield layer
#        queue.extend(layer.layers)
#        assert id(layer) not in seen # Catch loops!
#        seen.add(id(layer))
#
#
#def describe_all_params(model):
#    for layer in model.all_layers:
#        for name, shape, init in layer.describe_params:
#            yield name, shape, init
#
#



from .affine import Affine
from ... import describe
from ... import check
from ...describe import Dimension, Synapses, Biases
from ...check import has_shape


class ReLu(Affine):
    @check.arg(1, has_shape(('nB', 'nI')))
    def predict(self, input__BI):
        output__BO = Affine.predict(self, input__BI)
        output__BO = self.ops.xp.ascontiguousarray(output__BO, dtype='float32')
        self.ops.relu(output__BO, inplace=True)
        return output__BO

    @check.arg(1, has_shape(('nB', 'nI')))
    def begin_update(self, input__BI, drop=0.0):
        output__BO, finish_affine = Affine.begin_update(self, input__BI, drop=0.)
        output_copy = self.ops.xp.ascontiguousarray(output__BO, dtype='float32')
        self.ops.relu(output_copy, inplace=True)
        @check.arg(0, has_shape(('nB', 'nO')))
        def finish_update(gradient, sgd=None):
            gradient = self.ops.xp.ascontiguousarray(gradient, dtype='float32')
            self.ops.backprop_relu(gradient, output_copy, inplace=True)
            return finish_affine(gradient, sgd)
        output__BO[:] = output_copy
        output__BO, bp_dropout = self.ops.dropout(output__BO, drop, inplace=True)
        return output__BO, bp_dropout(finish_update)



# encoding: utf8
from __future__ import unicode_literals
import pytest
from hypothesis import given, strategies

from ...neural._classes import model as base
from ...neural.ops import NumpyOps


@pytest.fixture
def model_with_no_args():
    model = base.Model()
    return model


def test_Model_defaults_to_name_model(model_with_no_args):
    assert model_with_no_args.name == 'model'


def test_changing_instance_name_doesnt_change_class_name():
    model = base.Model()
    assert model.name != 'changed'
    model.name = 'changed'
    model2 = base.Model()
    assert model2.name != 'changed'


def test_changing_class_name_doesnt_change_default_instance_name():
    model = base.Model()
    assert model.name != 'changed'
    base.Model.name = 'changed'
    assert model.name != 'changed'
    # Reset state
    base.Model.name = 'model'

def test_changing_class_name_doesnt_changes_nondefault_instance_name():
    model = base.Model(name='nondefault')
    assert model.name == 'nondefault'
    base.Model.name = 'changed'
    assert model.name == 'nondefault'


def test_Model_defaults_to_cpu(model_with_no_args):
    assert isinstance(model_with_no_args.ops, NumpyOps)


def test_models_get_different_ids(model_with_no_args):
    model1 = base.Model()
    model2 = base.Model()
    assert model1.id != model2.id

def test_init_assigns_attributes():
    model = base.Model()
    model._mem
    assert model._layers == []
    assert model._operators == {}


def test_init_installs_via_descriptions():
    def mock_install(attr, self):
        setattr(self, attr, 'model=' + self.name)
    base.Model.descriptions = [("myattr", mock_install)]
    model = base.Model(name='model1')
    assert model.myattr == 'model=%s' % 'model1'
    model2 = base.Model(name='model2')
    assert model2.myattr == 'model=%s' % 'model2'


def test_init_calls_hooks():
    def mock_init_hook(self, *args, **kwargs):
        setattr(self, 'hooked', (args, kwargs))
    base.Model.on_init_hooks = [mock_init_hook]
    model = base.Model(0, 1, 2)
    assert model.hooked == ((0, 1, 2), {})
    model2 = base.Model(value='something')
    assert model2.hooked == (tuple(), {'value': 'something'})


def test_use_device():
    dev_id = id(base.Model.ops)
    with base.Model.use_device(base.Model.ops.device):
        assert id(base.Model.ops) == dev_id
    with base.Model.use_device('gpu'):
        assert id(base.Model.ops) != dev_id
    assert id(base.Model.ops) == dev_id


def test_bind_plus():
    with base.Model.define_operators({'+': lambda a, b: (a.name, b.name)}):
        m = base.Model(name='a') + base.Model(name='b')
        assert m == ('a', 'b')

def test_plus_chain():
    with base.Model.define_operators({'+': lambda a, b: a}):
        m = base.Model(name='a') + base.Model(name='b') + base.Model(name='c') + base.Model(name='d')
        assert m.name == 'a'


@pytest.mark.parametrize('op', '+ - * @ / // % ** << >> & ^ |'.split())
def test_all_operators(op):
    m1 = base.Model(name='a')
    m2 = base.Model(name='b')
    with base.Model.define_operators({op: lambda a, b: a.name + b.name}):
        if op == '+':
            value = m1 + m2
        else:
            with pytest.raises(TypeError):
                value = m1 + m2
        if op == '-':
            value = m1 - m2
        else:
            with pytest.raises(TypeError):
                value = m1 - m2

        if op == '*':
            value = m1 * m2
        else:
            with pytest.raises(TypeError):
                value = m1 * m2

        if op == '@':
            value = m1.__matmul__(m2) # Be kind to Python 2...
        else:
            with pytest.raises(TypeError):
                value = m1.__matmul__(m2)

        if op == '/':
            value = m1 / m2
        else:
            with pytest.raises(TypeError):
                value = m1 / m2

        if op == '//':
            value = m1 // m2
        else:
            with pytest.raises(TypeError):
                value = m1 // m2
        if op == '^':
            value = m1 ^ m2
        else:
            with pytest.raises(TypeError):
                value = m1 ^ m2
        if op == '%':
            value = m1 % m2
        else:
            with pytest.raises(TypeError):
                value = m1 % m2
        if op == '**':
            value = m1 ** m2
        else:
            with pytest.raises(TypeError):
                value = m1 ** m2
        if op == '<<':
            value = m1 << m2
        else:
            with pytest.raises(TypeError):
                value = m1 << m2
        if op == '>>':
            value = m1 >> m2
        else:
            with pytest.raises(TypeError):
                value = m1 >> m2
        if op == '&':
            value = m1 & m2
        else:
            with pytest.raises(TypeError):
                value = m1 & m2
        if op == '^':
            value = m1 ^ m2
        else:
            with pytest.raises(TypeError):
                value = m1 ^ m2
        if op == '|':
            value = m1 | m2
        else:
            with pytest.raises(TypeError):
                value = m1 | m2
    assert base.Model._operators == {}



import pytest
from ..neural.util import get_ops
from ..neural.ops import NumpyOps, CupyOps


def test_get_ops():
    ops = get_ops('numpy')
    assert isinstance(ops, NumpyOps)
    ops = get_ops('cpu')
    assert isinstance(ops, NumpyOps)
    ops = get_ops('cupy')
    assert isinstance(ops, CupyOps)
    ops = get_ops('gpu')
    assert isinstance(ops, CupyOps)
    with pytest.raises(ValueError):
        ops = get_ops('blah')


@pytest.mark.xfail
@pytest.mark.parametrize(
    'ids_batch', [
        (
            (
                ('the', 'cat', 'sat'),
                ('on', 'the', 'cat')
            ),
        ),
        (
            (
                ('the',),
                ('the',)
            ),
        ),
        (
            (
                ('a', 'b', 'a', 'b', 'd', 'e', 'f'),
            ),
        )
    ]
)
def test_get_positions(ids_batch):
    ids_batch = list(toolz.concat(ids_batch))
    positions = _get_positions(ids_batch)
    for key, idxs in positions.items():
        for i in idxs:
            assert ids_batch[i] == key


